'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var savUtil = require('sav-util');
var savRouter = require('sav-router');
var savSchema = require('sav-schema');

function getPayloads (route, {Vue, router, deepth}) {
  let components = router.getMatchedComponents(route);
  let arr = [];
  getComponentsDepth(Vue, components, deepth, arr);
  let payloads = [];
  arr.reduce(processComponent, payloads);
  return resolvePayloads(route, payloads)
}

function resolvePayloads (vueRoute, payloads) {
  return savUtil.clone(payloads.reduce((routes, payload) => {
    if (savUtil.isFunction(payload)) {
      payload = payload(vueRoute);
    }
    if (savUtil.isObject(payload)) { // 单个
      routes.push(payload);
    } else if (savUtil.isArray(payload)) { // 多个
      return routes.concat(payload.map(it => {
        if (savUtil.isString(it)) { // 简单字符串
          it = {name: it};
        }
        if (savUtil.isFunction(it)) {
          it = it(vueRoute);
        }
        return it
      }).filter(it => savUtil.isObject(it)))
    } else if (savUtil.isString(payload)) { // 简单字符串
      routes.push({name: payload});
    }
    return routes
  }, []))
}

function processComponent (payloads, component) {
  let options = typeof component === 'object' ? component : component.options;
  if (options.payload) {
    payloads.push(options.payload);
  }
  return payloads
}

function getComponentsDepth (Vue, components, depth, arr) {
  if (Array.isArray(components)) {
    for (let i = 0; i < components.length; ++i) {
      appendComponent(Vue, components[i], depth, arr);
    }
  } else {
    for (let comName in components) {
      appendComponent(Vue, components[comName], depth, arr);
    }
  }
}

function appendComponent (Vue, com, depth, arr) {
  if (savUtil.isString(com)) {
    com = Vue.component(com);
  }
  if (com) {
    arr.push(com);
    if (depth && com.components) {
      getComponentsDepth(Vue, com.components, depth--, arr);
    }
  }
}

const PAYLOAD_START = 'payload-start';
const PAYLOAD_PROGRESS = 'payload-progress';
const PAYLOAD_RESOLVE = 'payload-resolve';
const PAYLOAD_REJECT = 'payload-reject';

const REQUEST = 'request';
const RESPONSE = 'response';

const defaultProject = 'default';

class Sav {
  constructor (opts = {}) {
    this.opts = savUtil.testAssign(opts, {
      flux: null,
      contract: null,
      cacheField: true,
      cacheEnum: true
    });
    savUtil.bindEvent(this);
    this.contracts = {};
    this.caches = {};
    if (this.opts.contract) {
      this.setContract(this.opts.contract, defaultProject);
    }
  }
  /**
   * 添加新的Contract
   * @param {Contract} contract 合同
   * @param {String} name     别名
   */
  setContract (contract, name) {
    let isDefault = name === defaultProject;
    contract.injectFlux(this.opts.flux, isDefault);
    if (name) {
      this.contracts[name] = contract;
    }
    this.contracts[contract.projectName] = contract;
  }
  /**
   * 获取Schema字段文本描述
   * @param  {String} path    字段路径
   * @param  {String} project 项目名称
   * @return {String}         字段文本
   * @example
   *   getFieldTitle('ResAccountLogin.username')
   */
  getFieldTitle (path, project = defaultProject) {
    let caches = this.caches;
    if (this.opts.cacheField) {
      if (caches[path]) {
        return caches[path]
      }
    }
    let proj = this.contracts[project];
    let [structName, fieldName] = path.split('.');
    let ret;
    try {
      let struct = proj.schema.getSchema(structName);
      let field = struct.fieldByName(fieldName);
      ret = field.opts.title || `${project}.${path}`;
    } catch (err) {
      ret = project === defaultProject ? `${path}` : `${project}.${path}`;
    } finally {
      if (this.opts.cacheField) {
        caches[path] = ret;
      }
    }
    return ret
  }
  /**
   * 获取Schema枚举文本
   * @param  {String} path     字符串
   * @param  {String} enumKey  枚举项键
   * @param  {String} project  项目名称
   * @return {String}          枚举项文本
   * @example
   *   getEnumTitle('Sex.male')
   *   getEnumTitle('Sex', 'male')
   */
  getEnumTitle (path, enumKey, project = defaultProject) {
    if (!enumKey) {
      let arr = path.split('.');
      path = arr[0];
      enumKey = arr[1] || '';
    }
    let uri = `${project}.${path}.${enumKey}`;
    let caches = this.caches;
    if (this.opts.cacheEnum) {
      if (caches[uri]) {
        return caches[uri]
      }
    }
    let sav = this.contracts[project];
    let ret;
    try {
      let schemaEnum = sav.schema.getSchema(path);
      ret = schemaEnum.getEnum('key', enumKey).title;
    } catch (err) {
      ret = project === defaultProject ? `${path}.${enumKey}` : uri;
      if (enumKey === '') {
        ret = ret.substr(0, ret.length - 1);
      }
    } finally {
      if (this.opts.cacheEnum) {
        caches[uri] = ret;
      }
    }
    return ret
  }
  /**
   * 获取枚举项列表
   * @param  {String} enumName 枚举名称
   * @param  {String|Array} excludes 枚举项key过滤
   * @param  {String} project  项目名称
   * @return {Array}           枚举列表
   * @example
   *   getEnumItems('Sex')
   *   getEnumItems('Sex', 'male,unknown')
   *   getEnumItems('Sex', ['male'])
   */
  getEnumItems (enumName, excludes, project = defaultProject) {
    let proj = this.contracts[project];
    let ret;
    try {
      let schemaEnum = proj.schema.getSchema(enumName);
      ret = schemaEnum.opts.enums;
      if (savUtil.isString(excludes)) {
        excludes = excludes.split(',');
      }
      if (savUtil.isArray(excludes)) {
        return ret.filter(it => excludes.indexOf(it.key) === -1)
      }
    } catch (err) {
      ret = [];
    }
    return ret
  }
  /**
   * 预加载处理
   * @param  {Array} payloads 预加载
   * @param  {Object} route vue跳转路由
   * @return {Promise}
   */
  resolvePayloads (payloads, route) {
    let states = [];
    return new Promise((resolve, reject) => {
      try {
        payloads = payloads.filter((payload) => {
          let projectName = payload.project || defaultProject;
          let project = this.contracts[projectName];
          if (project) {
            if (project.resolvePayload(payload, route)) {
              if (payload.state) {
                states.push(payload.state);
              } else {
                return true
              }
            }
          } else {
            throw new Error('project no found: ', projectName)
          }
        });
      } catch (err) {
        return reject(err)
      }
      resolve({payloads, states});
    })
  }
  /**
   * 执行预加载
   * @param  {Array} payloads 预加载
   * @param  {Object} route    vue路由
   * @return {Promise}
   */
  invokePayloads (payloads, route) {
    return this.resolvePayloads(payloads, route)
      .then(({payloads, states}) => {
        if (payloads.length) {
          let status = {
            route,
            total: payloads.length,
            remains: payloads.length
          };
          this.emit(PAYLOAD_START, status);
          const progress = () => {
            status.remains--;
            this.emit(PAYLOAD_PROGRESS, status);
          };
          return Promise.all(payloads.map(async (payload) => {
            const p = payload.contract.invokePayload(payload);
            p.then(progress, progress);
            return p
          })).then((args) => {
            this.emit(PAYLOAD_RESOLVE, status);
            args = states.concat(args);
            if (args.length) {
              args = Object.assign.apply({}, args);
              return this.opts.flux.updateState(args)
            }
          }).catch((err) => {
            status.error = err;
            this.emit(PAYLOAD_REJECT, status);
            throw err
          })
        } else if (states.length) {
          return this.opts.flux.updateState(Object.assign.apply({}, states))
        }
      })
  }
}

class SavVue extends Sav {
  constructor (opts = {}) {
    super(savUtil.testAssign(opts, {
      fallback: null, // 请求失败后处理回调
      Vue: null,
      router: null
    }));
    let {router, Vue} = this.opts;

    router.beforeEach((to, from, next) => {
      let payloads = getPayloads(to, this.opts);
      this.invokePayloads(payloads, to)
        .then(() => {
          if (to.meta.title) {
            document.title = to.meta.title;
          }
          next();
        }, (err) => {
          let {fallback} = this.opts;
          if (fallback) {
            return fallback(to, from, next, err)
          }
          next(err);
        });
    });

    Vue.filter('fieldTitle', this.getFieldTitle.bind(this));
    Vue.filter('enumTitle', this.getEnumTitle.bind(this));
    Vue.filter('enumList', this.getEnumItems.bind(this));
  }
}

class Request {
  constructor (opts = {}) {
    this.opts = savUtil.testAssign(opts, {
      baseUrl: '/',
      stripHeaders: true, // 不返回头部信息
      ajax: savUtil.ajax
    });
    this.invokeQueues = [this.invoke.bind(this)];
    this.invoker = null;
    savUtil.bindEvent(this);
  }
  insert (fn) {
    this.invoker = null;
    this.invokeQueues.unshift(fn);
  }
  append (fn) {
    this.invoker = null;
    this.invokeQueues.push(fn);
  }
  request (options) {
    options = assign({
      url: '',
      dataType: 'JSON',
      headers: {
        Accept: 'application/json'
      }
    }, options);
    let {stripHeaders, baseUrl} = this.opts;
    if ('stripHeaders' in options) {
      stripHeaders = options.stripHeaders;
    }
    if (!/^(http(s?):)?\/\//i.test(options.url)) {
      options.url = baseUrl + options.url.replace(/^\//, '');
    }
    if (!this.invoker) {
      this.invoker = savUtil.compose(this.invokeQueues);
    }
    let {invoker} = this;
    let ctx = {request: options};
    return invoker(ctx).then(() => stripHeaders ? ctx.response.data : ctx.response)
  }
  invoke (ctx, next) {
    return new Promise((resolve, reject) => {
      ctx.xhr = this.opts.ajax(ctx.request, (err, data, headers) => {
        if (err) {
          return reject(err)
        }
        try {
          ctx.response = {
            data,
            headers
          };
          this.emit(RESPONSE, ctx);
        } catch (err) {
          return reject(err)
        }
        resolve();
      });
      this.emit(REQUEST, ctx);
    }).then(next)
  }
}

function assign (target, opts) {
  return Object.assign(target, opts)
}

/**
 * 接口数据缓存
 */
class Cache {
  constructor () {
    this.count = 0;
    this.caches = {};
  }
  get (key) {
    let val = this.caches[key];
    if (val) {
      let now = +new Date();
      if (now < val.now + val.ttl * 1000) {
        return JSON.parse(val.value)
      }
      delete this.caches[key];
      this.count--;
    }
  }
  set (key, ttl, name, value) {
    let exists = key in this.caches;
    this.caches[key] = {
      now: +new Date(),
      ttl,
      name,
      value: JSON.stringify(value)
    };
    if (!exists) {
      this.count++;
    }
  }
  clear () {
    this.caches = {};
    this.count = 0;
  }
  removeByName (name) {
    let it;
    let cacheKey;
    for (cacheKey in this.caches) {
      it = this.caches[cacheKey];
      if (it.name === name) {
        delete this.caches[cacheKey];
        this.count--;
      }
    }
  }
}

class Contract {
  constructor (opts = {}) {
    this.opts = savUtil.testAssign(opts, {
      enableMock: false, // 是否开启mock
      mockFlow: false,
      strict: true
    });
    let router = this.router = new savRouter.Router(this.opts);
    router.on('declareAction', (route) => {
      let name = route.name;
      this.routes[name] = route;
    });
    this.schema = new savSchema.Schema(this.opts);
    this.routes = {};
    this.mocks = {};
    this.request = new Request(this.opts);
    this.cache = new Cache();
    savUtil.bindEvent(this);
    if (this.opts.contract) {
      this.load(this.opts.contract);
    }
  }
  get projectName () {
    return this.project.name
  }
  /**
   * 加载contract
   * @param  {Object} data contract内容
   */
  load (data) {
    let {project, mocks} = data;
    this.schema.load(data);
    this.router.load(data);
    if (mocks) {
      mocks.forEach(mock => {
        if (mock.req) { // 不处理req, 只处理res
          return
        }
        let name = savUtil.pascalCase(mock.modalName + '_' + mock.actionName);
        let datas = this.mocks[name] || (this.mocks[name] = []);
        datas.push(mock);
      });
    }
    if (project) {
      this.project = project;
    }
  }
  /**
   * 处理payload
   * @param  {Object} payload  预加载配置
   * payload: {
   *   name: 'HomeSession', // Action路径 或 Schema名称
   *   merge: true, // 是否合并 vueRoute的query和params
   *   params: {}, // 自定义params
   *   query: {}, // 自定义query (payload只支持GET方法)
   *
   *   // 处理
   *   route, // 添加sav-route
   *   contract, // 添加Contract
   *   state, // 添加state
   * }
   * @param  {Object} vueRoute 当前vue路由
   */
  resolvePayload (payload, {params, query} = {}) {
    let savRoute = this.routes[payload.name];
    let schema = this.schema.getSchema(payload.name);
    if (savRoute) {
      if (payload.merge) {
        payload.params = Object.assign({}, params, payload.params);
        payload.query = Object.assign({}, query, payload.query);
      }
      savUtil.prop(payload, 'route', savRoute);
      savUtil.prop(payload, 'contract', this);
      return payload
    }
    if (schema) {
      let stateName = payload.name;
      let stateData;
      if (schema.schemaType === savSchema.Schema.SCHEMA_ENUM) {
        stateData = JSON.parse(JSON.stringify(schema.opts.enums));
      } else {
        stateData = schema.create(Object.assign({}, schema.opts.state));
        if (payload.merge) {
          Object.assign(stateData, params, query, payload.params, payload.query);
        }
      }
      payload.state = {
        [`${stateName}`]: stateData
      };
      return payload
    }
    let err = new Error('can not resolve payload');
    err.payload = payload;
    throw err
  }
  injectFlux (flux, isDefault) {
    let {schema} = this;
    let actions = Object.keys(this.routes).reduce((ret, routeName) => {
      let route = this.routes[routeName];
      let actionName = savUtil.pascalCase(route.method.toLowerCase() + '_' + (isDefault ? '' : this.projectName) + route.name);
      ret[actionName] = (flux, data, fetch) => {
        let argv = Object.assign({}, data);
        return this.invoke(flux, argv, route, fetch)
      };
      if (route.keys.length === 0) {
        ret[`${actionName}Data`] = (flux, data, fetch) => {
          let argv = {data};
          return this.invoke(flux, argv, route, fetch)
        };
      }
      let inputSchemaName = getRequestSchemaName(route);
      let reqStruct = schema.getSchema(inputSchemaName);
      if (reqStruct) {
        let reqName = (isDefault ? '' : this.projectName) + inputSchemaName;
        ret[`Check${reqName}`] = (flux, data) => {
          return reqStruct.check(data)
        };
        ret[`Extract${reqName}`] = (flux, data) => {
          return reqStruct.extractThen(data)
        };
      }
      let outputSchemaName = getResponseSchemaName(route);
      let resStruct = schema.getSchema(outputSchemaName);
      if (resStruct) {
        let resName = (isDefault ? '' : this.projectName) + outputSchemaName;
        ret[`Check${resName}`] = (flux, data) => {
          return resStruct.check(data)
        };
        ret[`Extract${resName}`] = (flux, data) => {
          return resStruct.extractThen(data)
        };
      }
      return ret
    }, {});
    flux.declare({
      actions
    });
  }
  invoke (flux, argv, route, fetch) {
    savUtil.prop(argv, 'route', route);
    return this.invokePayload(argv).then(async data => {
      if (fetch) {
        return data || argv.output
      }
      if (data) {
        await flux.updateState(data);
      }
    })
  }
  /**
   * 处理请求
   * @param  {Object} payload 请求上下文
   * payload: {
   *   params,
   *   query,
   *   ttl, // 缓存时效
   *   data, // 数据
   *   route: ... resolvePayload生成的路由
   *   // 生成内容
   *   url,    // 生成的请求地址 (缓存根据url和query生成)
   *   method, // 生成请求方法
   *   input,  // 生成输入参数 (用以校验)
   *   output, // 请求结果 (用以校验)
   * }
   * @step
   * 根据payload.route 和
   */
  async invokePayload (payload) {
    let {schema} = this;
    let {route} = payload;
    payload.url = route.compile(payload.params);
    let ttl = this.opts.noCache ? null : payload.ttl || (route.opts.ttl);
    let cacheKey = ttl ? getCacheKey(payload) : null;
    let cacheVal = cacheKey ? this.cache.get(cacheKey) : null;
    if (!cacheVal) {
      payload.input = Object.assign({}, payload.params, payload.query, payload.data);
      let inputSchema = schema.getSchema(getRequestSchemaName(route));
      if (inputSchema) {
        try {
          payload.input = inputSchema.extract(payload.input);
        } catch (err) {
          err.status = 400;
          throw err
        }
      }
      payload.method = route.method;
      let output = await this.fetch(payload);
      let outputSchema = schema.getSchema(getResponseSchemaName(route));
      let stateName;
      if (outputSchema) {
        stateName = outputSchema.name;
        outputSchema.check(output);
      }
      payload.output = output;
      payload.stateName = stateName;
      if (cacheKey) {
        this.cache.set(cacheKey, ttl, route.name, {output, stateName});
      }
    } else {
      payload.output = cacheVal.output;
      payload.stateName = cacheVal.stateName;
    }
    return mapPayloadState(payload)
  }
  fetch (payload) {
    if (this.opts.enableMock) {
      let mocks = this.mocks[payload.route.name];
      if (mocks && mocks.length) {
        if (this.opts.mockFlow) {
          return new Promise((resolve, reject) => {
            this.emit('mockFlow', {resolve, reject, payload, mocks});
          })
        } else {
          return mocks[0].data
        }
      } else {
        throw new Error(`mock data no found: ${payload.route.name}`)
      }
    }
    return this.request.request(payload)
  }
}

function getResponseSchemaName (route) {
  return route.opts.response || `Res${route.name}`
}

function getRequestSchemaName (route) {
  return route.opts.request || `Req${route.name}`
}

function getCacheKey (payload) {
  // 只支持query
  let uri = payload.url + JSON.stringify(payload.query);
  return savUtil.crc32(uri)
}

function getStatePath (output, stateName) {
  let pos = stateName.indexOf('.');
  while (pos !== -1) {
    output = output[stateName.substring(0, pos)];
    if (!savUtil.isObject(output)) {
      return
    }
    stateName = stateName.substr(pos + 1);
    pos = stateName.indexOf('.');
  }
  return output[stateName]
}

function mapPayloadState (payload) {
  let {route, output} = payload;
  let mapState = payload.mapState || route.opts.mapState;
  let ret = mapping(payload, mapState, output);
  let stateName = route.opts.stateName || payload.stateName;
  if (stateName) {
    return {[`${stateName}`]: ret}
  }
  return ret
}

function mapping (payload, mapState, output) {
  if (savUtil.isObject(output)) {
    if (savUtil.isArray(mapState)) {
      return mapState.reduce((ret, name) => {
        ret[name] = getStatePath(output, name);
        return ret
      }, {})
    } else if (savUtil.isObject(mapState)) {
      let ret = {};
      for (let name in mapState) {
        ret[name] = getStatePath(output, mapState[name]);
      }
      return ret
    }
  }
  if (savUtil.isFunction(mapState)) {
    return mapState(output)
  }
  return output
}

exports.SavVue = SavVue;
exports.Contract = Contract;
exports.PAYLOAD_START = PAYLOAD_START;
exports.PAYLOAD_PROGRESS = PAYLOAD_PROGRESS;
exports.PAYLOAD_RESOLVE = PAYLOAD_RESOLVE;
exports.PAYLOAD_REJECT = PAYLOAD_REJECT;
exports.REQUEST = REQUEST;
exports.RESPONSE = RESPONSE;
